import argparse
import pathlib
from typing import Union, List, Any
from typing import Optional as OptionalType
import dataclasses
import sys

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

from schema import Schema, Optional, Or, Use, SchemaError, And

from tclint.violations import Rule
from tclint import utils


@dataclasses.dataclass
class Config:
    """This dataclass defines the supported Config fields and their default
    values. It provides an external interface for accessing config values.

    The type annotations defined here are fairly loose - more specific type
    validation (and normalization) is defined by `validators` below.
    """

    exclude: List[Any] = dataclasses.field(default_factory=list)
    ignore: List[Any] = dataclasses.field(default_factory=list)
    commands: OptionalType[pathlib.Path] = dataclasses.field(default=None)
    extensions: List[str] = dataclasses.field(
        default_factory=lambda: ["tcl", "sdc", "xdc", "upf"]
    )
    style_indent: Union[str, int] = dataclasses.field(default=4)
    style_line_length: int = dataclasses.field(default=80)
    style_allow_aligned_sets: bool = dataclasses.field(default=False)
    style_max_blank_lines: int = dataclasses.field(default=2)
    style_indent_namespace_eval: bool = dataclasses.field(default=True)
    style_spaces_in_braces: bool = dataclasses.field(default=False)

    def apply_cli_args(self, args):
        args_dict = vars(args)
        for field in dataclasses.fields(self):
            if field.name in args_dict and args_dict[field.name] is not None:
                setattr(self, field.name, args_dict[field.name])

        # Special arguments that aren't handled automatically
        if args.extend_exclude is not None:
            self.exclude.extend(args.extend_exclude)

        if args.extend_ignore is not None:
            self.ignore.extend(args.extend_ignore)


# Validators using `schema` library that check and normalize config inputs.
# Used for checking both config files as well as config-related CLI args.

# Using these for CLI args adds a constraint that all non-boolean validators
# need to be able to normalize a value from a string. This means one could put
# e.g. a string representation of a list into a .toml config file, but we shouldn't
# document this, since it won't be considered stable behavior.

# _str2list handles this string-to-list normalization.


def _str2list(s):
    if isinstance(s, str):
        if s == "":
            return []
        return [v.strip() for v in s.split(",")]
    return s


_VALIDATORS = {
    # note: it's ok if paths don't exist - allows for generic
    # configurations with directories like .git/ excluded
    "exclude": Use(_str2list),
    "ignore": And(
        Use(_str2list),
        [
            Or(
                Use(
                    Rule,
                    error="invalid rule ID provided for 'ignore'",
                ),
                {
                    "path": Use(pathlib.Path),
                    "rules": [
                        Use(
                            Rule,
                            error="invalid rule ID provided for 'ignore'",
                        )
                    ],
                },
            )
        ],
    ),
    "commands": Use(pathlib.Path),
    "extensions": Use(_str2list),
    "style_indent": Or(
        lambda v: v == "tab", Use(int), error="indent must be integer or 'tab'"
    ),
    "style_line_length": Use(int, error="line-length must be integer"),
    "style_allow_aligned_sets": Use(bool, error="allow-aligned-sets must be a bool"),
    "style_max_blank_lines": And(
        Use(int),
        # we could technically support i >= 0, but I think 0 would be a weird
        # setting and this lets us ignore pluralizing the violation message :)
        lambda i: i >= 1,
        error="max-blank-lines must be an integer with value at least 1",
    ),
    "style_indent_namespace_eval": Use(
        bool, error="indent-namespace-eval must be a bool"
    ),
    "style_spaces_in_braces": Use(bool, error="spaces-in-braces must be a bool"),
}


def _validate_config(config):
    """Validates dictionary read from TOML config file. Individual value validators
    are implemented in the global dict, this defines the actual structure of the
    schema."""

    base_config = {
        Optional("ignore"): _VALIDATORS["ignore"],
        Optional("commands"): _VALIDATORS["commands"],
        Optional("style"): {
            Optional("indent"): _VALIDATORS["style_indent"],
            Optional("line-length"): _VALIDATORS["style_line_length"],
            Optional("allow-aligned-sets"): _VALIDATORS["style_allow_aligned_sets"],
            Optional("max-blank-lines"): _VALIDATORS["style_max_blank_lines"],
            Optional("indent-namespace-eval"): _VALIDATORS[
                "style_indent_namespace_eval"
            ],
            Optional("spaces-in-braces"): _VALIDATORS["style_spaces_in_braces"],
        },
    }

    schema = Schema({
        # exclude and extensions can only be used in global context
        Optional("exclude"): _VALIDATORS["exclude"],
        Optional("extensions"): _VALIDATORS["extensions"],
        **base_config,
        Optional("fileset"): Schema([{"paths": [Use(pathlib.Path)], **base_config}]),
    })

    try:
        return schema.validate(config)
    except SchemaError as e:
        if e.errors[-1] is not None:
            error = e.errors[-1]
        else:
            error = e.autos[-1]
        raise ConfigError(error)


def setup_config_cli_args(parser):
    """This method defines config-related CLI arguments.

    The destvars of these switches should match the fields of Config.
    """
    config_group = parser.add_argument_group("configuration arguments")

    def validator(key):
        def func(s):
            try:
                return Schema(_VALIDATORS[key]).validate(s)
            except SchemaError as e:
                error_s = str(e).replace("\n", " ")
                raise argparse.ArgumentTypeError(error_s)

        return func

    def add_bool(dest, yes_flag, no_flag):
        mutex_group = config_group.add_mutually_exclusive_group(required=False)
        mutex_group.add_argument(yes_flag, dest=dest, action="store_true")
        mutex_group.add_argument(no_flag, dest=dest, action="store_false")
        parser.set_defaults(**{dest: None})

    config_group.add_argument(
        "--ignore", type=validator("ignore"), metavar='"rule1, rule2, ..."'
    )
    config_group.add_argument(
        "--extend-ignore", type=validator("ignore"), metavar='"rule1, rule2, ..."'
    )
    config_group.add_argument(
        "--exclude", type=validator("exclude"), metavar='"path1, path2, ..."'
    )
    config_group.add_argument(
        "--extend-exclude", type=validator("exclude"), metavar='"path1, path2, ..."'
    )
    config_group.add_argument(
        "--extensions", type=validator("extensions"), metavar='"tcl, xdc, ..."'
    )
    config_group.add_argument(
        "--commands", type=validator("commands"), metavar="<path>"
    )
    config_group.add_argument(
        "--style-indent", type=validator("style_indent"), metavar="<indent>"
    )
    config_group.add_argument(
        "--style-line-length",
        type=validator("style_line_length"),
        metavar="<line_length>",
    )
    config_group.add_argument(
        "--style-max-blank-lines",
        type=validator("style_max_blank_lines"),
        metavar="<max_blank_lines>",
    )

    add_bool(
        "style_allow_aligned_sets",
        "--style-aligned-sets",
        "--style-no-aligned-sets",
    )
    add_bool(
        "style_indent_namespace_eval",
        "--style-indent-namespace-eval",
        "--style-no-indent-namespace-eval",
    )
    add_bool(
        "style_spaces_in_braces",
        "--style-spaces-in-braces",
        "--style-no-spaces-in-braces",
    )


def _flatten(d, prefix=None):
    """Flattens TOML config dictionary structure to match the flat set of fields
    expected by Config dataclass."""
    if prefix is None:
        prefix = []

    flat = {}
    for k, v in d.items():
        if isinstance(v, dict):
            flat.update(_flatten(v, prefix=prefix + [k]))
        else:
            flat["_".join(prefix + [k]).replace("-", "_")] = v

    return flat


class RunConfig:
    """Class that holds information about both global and fileset configs. User
    code can get a Config object that applies to a particular file by calling
    get_from_path() and supplying that file's path."""

    def __init__(self, global_config=None, fileset_configs=None):
        if global_config is not None:
            self._global_config = global_config
        else:
            self._global_config = Config()

        self._fileset_configs = [
            # ([pathlib.Path...], Config])
        ]
        if fileset_configs is not None:
            self._fileset_configs = fileset_configs

    @property
    def exclude(self):
        return self._global_config.exclude

    @property
    def extensions(self):
        return self._global_config.extensions

    @classmethod
    def from_dict(cls, config_dict: dict):
        config_dict = _validate_config(config_dict)
        try:
            fileset_config_dicts = config_dict.pop("fileset")
        except KeyError:
            fileset_config_dicts = []

        config_dict = _flatten(config_dict)
        global_config = Config(**config_dict)

        fileset_configs = []
        for fileset_config in fileset_config_dicts:
            paths = fileset_config.pop("paths")

            paths = tuple([path.resolve() for path in paths])
            fileset_config = _flatten(fileset_config)

            # pull in default values from global config
            full_fileset_config = config_dict.copy()
            full_fileset_config.update(fileset_config)

            fileset_configs.append((paths, Config(**full_fileset_config)))

        return cls(global_config, fileset_configs)

    @classmethod
    def from_path(cls, path: Union[str, pathlib.Path]):
        path = pathlib.Path(path)

        if not path.exists():
            raise FileNotFoundError

        with open(path, "rb") as f:
            try:
                data = tomllib.load(f)
            except tomllib.TOMLDecodeError as e:
                raise ConfigError(f"{path}: {e}")

        try:
            return cls.from_dict(data)
        except ConfigError as e:
            raise ConfigError(f"{path}: {e}")

    @classmethod
    def from_pyproject(cls, directory=None):
        if directory is None:
            directory = pathlib.Path(".")
        else:
            directory = pathlib.Path(directory)

        path = directory / "pyproject.toml"

        if not path.exists():
            raise FileNotFoundError

        with open(path, "rb") as f:
            data = tomllib.load(f)

        tclint_config = data.get("tool", {})["tclint"]

        try:
            return cls.from_dict(tclint_config)
        except ConfigError as e:
            raise ConfigError(f"pyproject.toml: {e}")

    def get_for_path(self, path) -> Config:
        if path is None:
            return self._global_config

        path = path.resolve()
        for fileset_paths, config in self._fileset_configs:
            for fileset_path in fileset_paths:
                if utils.is_relative_to(path, fileset_path):
                    return config

        return self._global_config

    def apply_cli_args(self, args):
        self._global_config.apply_cli_args(args)
        for _, fileset_config in self._fileset_configs:
            fileset_config.apply_cli_args(args)


class ConfigError(Exception):
    pass


def get_config(config_path) -> RunConfig:
    DEFAULT_CONFIGS = ("tclint.toml", ".tclint")

    # user-supplied
    if config_path is not None:
        try:
            return RunConfig.from_path(config_path)
        except FileNotFoundError:
            raise ConfigError(f"path {config_path} doesn't exist")

    for path in DEFAULT_CONFIGS:
        try:
            return RunConfig.from_path(path)
        except FileNotFoundError:
            pass

    try:
        return RunConfig.from_pyproject()
    except ConfigError as e:
        raise e
    except (FileNotFoundError, tomllib.TOMLDecodeError, KeyError):
        # just skip if file doesn't exist, contains TOML errors, or tclint key not found
        pass

    return RunConfig()



import os
import pathlib
import subprocess

import pytest

from tclint import tclint

MY_DIR = pathlib.Path(__file__).parent.resolve()

tests = []
test_case_dir = MY_DIR / "data"
for path in test_case_dir.iterdir():
    if path.suffix == ".tcl":
        tests.append(str(path.relative_to(MY_DIR)))


@pytest.mark.parametrize("test", tests)
def test_tclint(test):
    """End-to-end tests."""
    output_path = (MY_DIR / test).with_suffix(".txt")
    with open(output_path, "r") as f:
        expected = f.read()

    cmd = ["tclint", test]

    config_file = (MY_DIR / test).with_suffix(".toml")
    if config_file.exists():
        cmd += ["-c", config_file]

    p = subprocess.run(cmd, capture_output=True, cwd=MY_DIR)

    assert p.stdout.decode("utf-8") == expected
    assert p.returncode == 0 if not expected else 1


def test_tclint_show_categories():
    test = (test_case_dir / "example.tcl").relative_to(MY_DIR)

    p = subprocess.run(
        [
            "tclint",
            "--show-categories",
            test,
        ],
        capture_output=True,
        cwd=MY_DIR,
    )
    expected = """
data/example.tcl:1:1: too many args for puts: got 4, expected no more than 3 [func:command-args]
data/example.tcl:2:1: expected indent of 0 spaces, got 2 [style:indent]
data/example.tcl:3:5: expected 1 space between words, got 3 [style:spacing]
""".lstrip()  # noqa E501

    assert p.stdout.decode("utf-8") == expected
    assert p.returncode == 1


def test_switches():
    test = (test_case_dir / "dirty.tcl").relative_to(MY_DIR)
    config_file = test_case_dir / "tclint.toml"

    p = subprocess.run(
        [
            "tclint",
            "-c",
            config_file,
            "--style-line-length",
            "35",
            "--ignore",
            "",
            test,
        ],
        capture_output=True,
        cwd=MY_DIR,
    )

    expected = """
data/dirty.tcl:1:14: expected 1 space between words, got 2 [spacing]
data/dirty.tcl:1:36: line length is 36, maximum allowed is 35 [line-length]
data/dirty.tcl:2:1: expected indent of 2 spaces, got 0 [indent]
data/dirty.tcl:3:1: expected indent of 4 spaces, got 0 [indent]
data/dirty.tcl:4:1: expected indent of 2 spaces, got 0 [indent]
data/dirty.tcl:4:2: expected 1 space between words, got 3 [spacing]
data/dirty.tcl:5:1: expected indent of 4 spaces, got 8 [indent]
data/dirty.tcl:6:1: expected indent of 2 spaces, got 0 [indent]
data/dirty.tcl:6:17: expression with substitutions should be enclosed by braces [unbraced-expr]
data/dirty.tcl:7:1: expected indent of 4 spaces, got 8 [indent]
data/dirty.tcl:7:13: expected 1 space between words, got 3 [spacing]
data/dirty.tcl:8:1: expected indent of 2 spaces, got 4 [indent]
data/dirty.tcl:9:1: expected indent of 4 spaces, got 8 [indent]
data/dirty.tcl:9:13: expected 1 space between words, got 5 [spacing]
data/dirty.tcl:10:1: expected indent of 2 spaces, got 4 [indent]
""".lstrip()  # noqa E501

    assert p.stdout.decode("utf-8") == expected
    assert p.returncode == 1


def test_special_file():
    p = subprocess.run(["tclint", "/dev/stdin"])
    assert p.returncode == 0


def test_read_stdin():
    p = subprocess.Popen(
        ["tclint", "-"],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = p.communicate(input=' puts "hello world"'.encode("utf-8"))

    output = stdout.decode("utf-8").strip()
    assert output == "(stdin):1:1: expected indent of 0 spaces, got 1 [indent]"
    assert stderr == b""


def test_resolve_sources(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("a")
    (tmp_path / "src").mkdir()
    (tmp_path / "src" / "ignore").mkdir()
    (tmp_path / "src" / "ignore" / "bar.tcl").touch()
    (tmp_path / "src" / "ignore1.tcl").touch()
    (tmp_path / "src" / "ignore2.tcl").touch()
    to_include = tmp_path / "src" / "foo.tcl"
    to_include.touch()

    cwd = os.getcwd()
    os.chdir(tmp_path)

    extensions = ["tcl"]

    sources = tclint.resolve_sources(
        [pathlib.Path(".")],
        exclude_patterns=[
            # test bare pattern matches anywhere in tree
            "ignore",
            # test glob
            "ignore*.tcl",
            # test pattern with "/" only matches from root of tree
            "/foo.tcl",
        ],
        exclude_root=tmp_path,
        extensions=extensions,
    )

    assert len(sources) == 1
    assert sources[0] == to_include.relative_to(tmp_path)

    # test absolute path outside of exclude_root doesn't get matched by ignore
    # pattern starting with "/"
    other_dir = tmp_path_factory.mktemp("b")
    in_path = other_dir / "foo.tcl"
    in_path.touch()
    sources = tclint.resolve_sources(
        [in_path],
        exclude_patterns=[str(in_path)],
        exclude_root=tmp_path,
        extensions=extensions,
    )
    assert len(sources) == 1
    assert sources[0] == in_path

    # test that we can match outside of exclude root with explicit relative path
    top_src = tmp_path / "top.tcl"
    top_src.touch()
    sources = tclint.resolve_sources(
        [top_src],
        exclude_patterns=["../top.tcl"],
        exclude_root=tmp_path / "src",
        extensions=extensions,
    )
    assert len(sources) == 0

    # test auto-escape leading hash
    other_other_dir = tmp_path_factory.mktemp("c")
    hash_srcs = [other_other_dir / "#foo.tcl", other_other_dir / "#bar.tcl"]
    for src in hash_srcs:
        src.touch()
    sources = tclint.resolve_sources(
        hash_srcs,
        # extra space before #bar.tcl is important to make sure we don't just match ^#
        exclude_patterns=["#foo.tcl", " #bar.tcl"],
        exclude_root=other_other_dir,
        extensions=extensions,
    )
    assert len(sources) == 0

    os.chdir(cwd)


def test_resolve_sources_extensions(tmp_path):
    foo_file = tmp_path / "file.foo"
    foo_file.touch()
    bar_file = tmp_path / "file.BAR"
    bar_file.touch()

    cwd = os.getcwd()
    os.chdir(tmp_path)

    sources = tclint.resolve_sources(
        [tmp_path], exclude_patterns=[], exclude_root=tmp_path, extensions=["foo"]
    )
    assert len(sources) == 1
    assert sources[0] == foo_file

    sources = tclint.resolve_sources(
        [tmp_path], exclude_patterns=[], exclude_root=tmp_path, extensions=["bar"]
    )
    assert len(sources) == 1
    assert sources[0] == bar_file

    os.chdir(cwd)



import argparse
import json
import shlex
import sys

from tclint.commands.plugins import PluginManager

try:
    from tclint._version import __version__  # type: ignore
except ModuleNotFoundError:
    __version__ = "(unknown version)"


def make_cmd_spec(args) -> int:
    plugin_name = args.plugin
    plugin = PluginManager.get_mod(plugin_name)
    if plugin is None:
        print(f"Unable to make command spec for {plugin_name}")
        return 1

    if not hasattr(plugin, "make_command_spec"):
        print(f"Plugin {plugin_name} does not support command spec generation")
        return 1

    exec_cmd = None
    if args.exec is not None:
        exec_cmd = shlex.split(args.exec)

    make_command_spec = getattr(plugin, "make_command_spec")
    try:
        command_spec = make_command_spec(exec_cmd)
    except Exception as e:
        print(f"Error encountered while generating command spec: {e}")
        return 1

    output_path = f"{args.plugin}.json" if args.output is None else args.output
    with open(output_path, "w") as out_file:
        json.dump({"plugin": plugin_name, "spec": command_spec}, out_file)
        print(f"Wrote command spec to {output_path}")

    return 0


def main() -> int:
    parser = argparse.ArgumentParser(
        "tclint-plugins", description="tclint plugin manager"
    )
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )
    subparsers = parser.add_subparsers(required=True)

    cmd_spec_parser = subparsers.add_parser(
        "make-spec", help="generate command spec from installed tool"
    )
    cmd_spec_parser.set_defaults(func=make_cmd_spec)
    cmd_spec_parser.add_argument("plugin", help="name of plugin to generate spec for")
    cmd_spec_parser.add_argument(
        "-o", "--output", metavar="<path>", help="path to store command spec"
    )
    cmd_spec_parser.add_argument(
        "-e",
        "--exec",
        metavar="<cmd>",
        help="path to binary or shell invocation for tool",
    )

    args = parser.parse_args()
    return args.func(args)


if __name__ == "__main__":
    sys.exit(main())



from pathlib import Path

from tclint.tclint import lint
from tclint.config import Config
from tclint.violations import Rule


def test_cmd_args_in_sub():
    """Ensure command args get checked in cmd sub."""
    script = "[puts]"
    violations = lint(script, Config(), Path())

    assert len(violations) == 1
    assert violations[0].id == Rule("command-args")


def test_no_false_positive_arg_expansion():
    """Ensure that argument count checks don't flag when the number of arguments
    is ambiguous."""
    # $foo may be a list with two items, which is legal
    script = "rename {*}$foo"
    violations = lint(script, Config(), Path())
    assert len(violations) == 0


def test_aligned_sets():
    """Test style.allow-aligned-sets"""
    script = """
set foo  0
set barx 1"""

    violations = lint(script, Config(style_allow_aligned_sets=False), Path())
    assert len(violations) == 1
    assert violations[0].id == Rule("spacing")

    violations = lint(script, Config(style_allow_aligned_sets=True), Path())
    assert len(violations) == 0


def test_accidental_return_expr():
    script = "return 5 + 2"
    violations = lint(script, Config(), Path())
    assert len(violations) == 1
    assert violations[0].id == Rule("command-args")


def test_ignore_path():
    script = "puts  bad_spacing"
    fake_path = Path("bad.tcl")
    violations = lint(script, Config(), fake_path)
    assert len(violations) == 1

    config = Config(ignore=[{"path": fake_path, "rules": [Rule("spacing")]}])
    violations = lint(script, config, fake_path)
    assert len(violations) == 0


def test_redefined_builtin():
    script = "proc puts {} {}"
    violations = lint(script, Config(), Path())
    assert len(violations) == 1
    assert violations[0].id == Rule("redefined-builtin")


def test_no_indent_namespace_eval():
    script = """
namespace eval my_namespace {
puts "okay no indent here"
 puts "indent errors still caught though"
if {1} {
puts "including in blocks"
}
}
"""
    violations = lint(script, Config(style_indent_namespace_eval=False), Path())
    assert len(violations) == 2
    assert violations[0].id == Rule.INDENT
    assert violations[0].pos[0] == 4
    assert violations[1].id == Rule.INDENT
    assert violations[1].pos[0] == 6


def test_no_violation_multiline_expr():
    script = r"""
if {1 && \
    2} {}"""
    violations = lint(script, Config(), Path())
    assert len(violations) == 0


def test_spacing_multiline():
    script = r"""
foo asdf asdf \
    qwerty    qwerty"""

    violations = lint(script, Config(), Path())
    assert len(violations) == 1
    assert violations[0].id == Rule.SPACING
    assert violations[0].pos == (3, 11)


def test_spacing_multiline_expr():
    script = r"""
expr {$a ?
      $b  : $d}"""

    violations = lint(script, Config(), Path())
    assert len(violations) == 1
    assert violations[0].id == Rule.EXPR_FORMAT
    assert violations[0].pos == (3, 9)
